﻿
using System;
using System.Drawing;
using System.Windows.Forms;

/*    4. order Runge Kutta algorithm with adaptive step size   */
/*                      www.mosismath.com                      */

namespace WindowsFormsApplication1
{
    public delegate void function();

    public partial class MainWin : Form
    {
        public const int datapoints = 400; // number of datapoints for the intergration
        public const double hMax = 0.5;
        public const double hMin = 1E-2;

        double[] b = new double[datapoints];
        double[] yp = new double[datapoints];
        double[] xp = new double[datapoints];
        double x = 0;
        double y = 1;

        public function func;


        double f_xy(double x, double y)
        {
            return -(Math.Sin(x*x*x)+3*x*x*x*Math.Cos(x*x*x))*y;
        }
        

        double Interpolate(double xp, int order)
        {
            int i, k;
            double tempYp = 0;
            double yp = 0;
            for (i = 1; i < order; i++)
            {
                tempYp = b[i];
                for (k = 0; k < i; k++)
                {
                    tempYp = tempYp * (xp);
                }
                yp = yp + tempYp;
            }
            return b[0] + yp;
        }
        

        public MainWin()
        {     
            InitializeComponent();
            func = CalcRungeKutta4;
        }


        public void DrawGraph(int maxPoints, double minX, double maxX, bool doClear)
        {
            Point p1, p2;
            int j;
            double maxTime = -1.0;
            double maxVal = -1.0;
            double minVal = 1.0;
            double scalefactor;
            bool drawCeroline;
            for (j = 0; j < maxPoints - 1; j++)
            {
                if ((xp[j] >= minX) && (xp[j] <= maxX))
                {
                    if (maxTime < xp[j])
                        maxTime = xp[j];
                    if (maxVal < yp[j])
                        maxVal = yp[j];
                    if (minVal > yp[j])
                        minVal = yp[j];
                }
            }
            maxTime = maxTime * 1.05;
            if (minVal < 0)
            {
                if (maxVal > 0)
                {
                    if (maxVal > Math.Abs(minVal))
                        scalefactor = pGraph.Height / maxVal / 2.2;
                    else
                        scalefactor = pGraph.Height / Math.Abs(minVal) / 2.2;
                }
                else
                    scalefactor = 5.0;
                drawCeroline = true;
            }
            else
            {
                scalefactor = pGraph.Height / (maxVal) / 1.1;
                drawCeroline = false;
            }
            p1 = new Point();
            p2 = new Point();
            Graphics g = pGraph.CreateGraphics();
            if (doClear)
                g.Clear(Color.White);
            Pen bluePen = new Pen(Color.Blue, 2);
            Pen redPen = new Pen(Color.Red, 2);
            Pen blackPen = new Pen(Color.Black, 2);
            if (drawCeroline)
            {
                p1.X = 0;
                p1.Y = (pGraph.Height / 2);
                p2.X = pGraph.Width;
                p2.Y = (pGraph.Height / 2);
                g.DrawLine(blackPen, p1, p2);
            }
            bluePen.Width = 1;
            if (maxTime > 0) // draw interpolated graph
            {
                for (j = 0; j < maxPoints - 1; j++)
                {
                    if ((xp[j] >= minX) && (xp[j] <= maxX))
                    {
                        p1.X = Convert.ToInt32(xp[j] * pGraph.Width / (double)maxTime);
                        p2.X = Convert.ToInt32(xp[j + 1] * pGraph.Width / (double)maxTime);
                        if (drawCeroline)
                        {
                            p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j] * scalefactor));
                            p2.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor));
                        }
                        else
                        {
                            p1.Y = pGraph.Height - Convert.ToInt32(Math.Round(yp[j] * scalefactor));
                            p2.Y = pGraph.Height - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor));
                        }
                        if(p1.X < p2.X)
                            g.DrawLine(bluePen, p1, p2);
                    }
                }
            }
            g.Dispose();
        }


        private void CalcRungeKutta4()
        {
            int i;
            double h = 3.0 / (double)(datapoints - 1);
            double k1, k2, k3, k4;
            // initial values
            x = 0;
            y = 1;
            for (i = 1; i < datapoints; i++)
            {
                k1 = f_xy(x, y);
                k2 = f_xy(x + h / 2, y + h * k1 / 2);
                k3 = f_xy(x + h / 2, y + h * (k1 + k2) / 4);
                k4 = f_xy(x + h, y + h * (-k2 + 2 * k3));
                y = y + h / 6 * (k1 + 4 * k3 + k4);
                x += h;
                xp[i] = x;
                yp[i] = y;
            }
        }

        private void CalcRungeKuttaVar()
        {
            int i=1;
            double h = hMax;
            double hp = 3.0 / (double)(datapoints - 1);
            double k1, k2, k3, k4;
            double t = 0;
            int count = 0;
            bool updateValues = true;

            for(i = 0; i < datapoints; i++)
            {
                xp[i] = 0;
                yp[i] = 0;
            }
            i = 1;
            // initial values
            x = 0;
            y = 1;
            xp[0] = x;
            yp[0] = y; 
            while (x < 3)
            {            
                k1 = f_xy(x, y);
                k2 = f_xy(x + h / 2, y + h / 2 * k1);
                k3 = f_xy(x + h / 2, y + h * (k1 + k2) / 4);
                k4 = f_xy(x + h, y + h * (-k2 + 2 * k3));                
                t = h / 6 * (k1 + 4 * k3 + k4 - 6 * k2);
                updateValues = true;
                if (Math.Abs(t) > 0.1)
                {
                    if (h > hMin)
                    {
                        h = h / 2;
                        updateValues = false;
                    }
                }
                else
                {
                    if (Math.Abs(t) < 1E-10)
                    {
                        if (h < hMax)
                        {
                            h = h * 2;
                            updateValues = false;
                        }
                    }
                }
                if(updateValues)
                {
                    count++;
                    y = y + h / 6 * (k1 + 4 * k3 + k4);
                    x += h;
                    if ((x - (i * hp) > 0) && (i < datapoints))
                    {
                        xp[i] = x;
                        yp[i] = y;
                        i++;
                    }
                }
            }
        }


        private void button1_MouseClick(object sender, MouseEventArgs e)
        {
            func?.Invoke(); //Call delegate
            DrawGraph(datapoints, 0, 3, true);
            tbX.Text = Math.Round(x, 4).ToString();
            tbY.Text = Math.Round(y, 8).ToString();
        }

        private void pGraph_Paint(object sender, PaintEventArgs e)
        {
            DrawGraph(datapoints, 0, 3, false);
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (comboBox1.SelectedIndex)
            {
                case 0:
                    func = CalcRungeKutta4;
                    break;
                case 1:
                    func = CalcRungeKuttaVar;
                    break;
            }
        }
    }
}
